/* Un outil pour produire des listes de clauses  hasard */
/* A utiliser pour tester votre projet */
/* Problme : comme amliorer le code ? */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#define TAILLE_INITIALE 1024 /* Du flot */
#define  MAX_CL 1000
#define  MAX_LIT 1000

static char * conjonction = ",";
static char * disjonction = "+";
static char * negation = "!";
static char * openens = "{";
static char * closeens = "}";
static char * openclause = "(";
static char * closeclause = ")";


static struct flot {
  char *tampon;
  char *curr;
  unsigned int taille;
} flot;


/* Engendrer chaines de caracteres */
void ens_engendrer(int,int);
void clause_engendrer(int);
void lit_engendrer();
/* Utilisation */
void usage (char *command);
/* Ecriture dans le flot */
int ecrire(struct flot *fl, const char *chaine);
void ajouter(char *chaine);
/* Traitement bas niveu du flot */
int initialiser(struct flot *fl);
int mettreajour(struct flot *fl);

int main(int argc, char *argv[])
{
  int max_noclauses=MAX_CL;
  int max_nolitsperclause=MAX_LIT;
  
  /* Gestion des options */
  int verbose=0;
  {
    char c;
    extern int optind;
    int noargs;
    
    while((c=getopt(argc,argv,"vp")) != -1)
      {
	switch( c){
	case 'v':
	  verbose=1;
	  break;
	case 'p':
	  conjonction = ",\n";
	  disjonction = "+";
	  negation = "!";
	  openens = "{\n";
	  closeens = "\n}";
	  openclause = "\t(";
	  closeclause = ")";
	  break;
	}
      }
    noargs = argc -optind;
    
    if(noargs != 2) usage(argv[0]);
    
    if(
       (sscanf(argv[optind],"%d",&max_noclauses) == 0)) 
      usage(argv[0]);
    
    if( 
       (sscanf(argv[optind + 1],"%d",&max_nolitsperclause) == 0)) 
      usage(argv[0]);
    
  }
  
  /* Initialisation des variables, 
     de la profondeur, et du gnrateur alatoire */ 

  initialiser(&flot);
  max_noclauses = (max_noclauses <= MAX_CL)?max_noclauses:MAX_CL;
  max_nolitsperclause = (max_nolitsperclause <= MAX_LIT)?max_nolitsperclause:MAX_LIT;
  srand( time(NULL) * getpid());
  
  /* Partie principale du programme */
  ens_engendrer(max_noclauses,max_nolitsperclause);
  
  /* Affichage */
  printf("%s\n",flot.tampon);
  if(verbose)
    printf("\nLonguer de la chaine : %d\n", strlen(flot.tampon));
  
  /* Sortie */
  free(flot.tampon);
  exit(EXIT_SUCCESS);
  
}

void ens_engendrer(int max_clauses,int max_lits)
{
  int i, noclauses;
    
  noclauses = rand()% (max_clauses + 1);
  ajouter(openens);
  for(i=0;i<noclauses;i++)
    {
      clause_engendrer(max_lits);
      if(i != (noclauses -1) )
	ajouter(conjonction);
    }
  ajouter(closeens);
} 

void clause_engendrer(int max_lits)
{
  int i, nolits;
   
  nolits = rand()% (max_lits + 1);
  ajouter(openclause);
  for(i=0;i<nolits;i++)
    {
      lit_engendrer();
      if(i != (nolits -1) )
	ajouter(disjonction);
    }
  ajouter(closeclause);
} 

void lit_engendrer()
{
  char str[3];
  char prop = 'a' + (rand()% ('z'-'a' + 1));
  char signe = rand()%2;
  
  if(signe)
    sprintf(str,"%c",prop);
  else
    sprintf(str, "!%c",prop);
  ajouter(str);
} 

void usage (char *command)
{
  printf(
	 "Usage : %s [-vp] max_clauses max_lits_per_clause\n"
	 "\t-v : verbose\n"
	 "\t-p : pretty\n"
 ,command);

  exit(EXIT_FAILURE);
}

int initialiser(struct flot *fl)
{
  if(fl == NULL)
    return -1;

  fl->tampon=calloc(TAILLE_INITIALE,sizeof(char));
  if(fl->tampon == NULL)
    exit(EXIT_FAILURE);
  fl->curr=fl->tampon;
  fl->taille=TAILLE_INITIALE;
  
  return 0;
}

int mettreajour(struct flot *fl)
{
  char *aux;

  aux = realloc(fl->tampon,fl->taille*2);
  if(aux == NULL)
    { 
      perror("realloc");
      return -1;
    }
  fl->curr = aux + (fl->curr - fl->tampon);
  fl->taille*=2;
  if(aux != fl->tampon)
    { 
      /*free(fl->tampon);*/ /* Voir l'implementation, avec cela ne marche pas */ 
      fl->tampon = aux ;
    }
  return 0;
}

int ecrire(struct flot *fl, const char *chaine)
{
  int n = strlen(chaine);
  
  if(fl->curr + n + 1 > fl->tampon + fl->taille)
    if( mettreajour(fl) == -1 ) return -1;
  
  strcpy(fl->curr,chaine);
  fl->curr+=n;
  
  return 0;
}

void ajouter(char *chaine)
{
  if( ecrire(&flot,chaine) == -1)
    {
      perror("Ecriture dans le flot");
      exit(EXIT_FAILURE);
    }
}
