(** Les types : formules atomiques, clauses, thories **)
type ('a,'b,'c) atomic_formula = Formula of 'a * ('b,'c) Terme.t list;;
type ('a,'b,'c) clause = { g : ('a,'b,'c) atomic_formula list ;
			   d : ('a,'b,'c) atomic_formula list 
			 };;
type ('a,'b,'c) theorie = ('a,'b,'c) clause list;;

(** Formules atomiques **)
let atf_faire p termes = Formula(p,termes);;
(* Printing, parsing, etc ... *)
let atf_to_string  = function 
    Formula(p,termes)->
      p^"("^(Outils.list_to_string "," Terme.to_string termes)^")";;
let atf_print atf = print_string(atf_to_string atf);;
let atf_parse = parser
      [<'Genlex.Ident p ; 'Genlex.Kwd "(";
	termes = Lexparse.parse_list "," Terme.parse; 'Genlex.Kwd ")">] ->
	atf_faire p termes ;;
let atf_of_string str = atf_parse(Lexparse.lexer str);;


(* Clauses  -- ici on ne met pas de prefixes*)
let vide = { g = []; d =[]};;
let faire g d = { g=g; d=d};;
let gauche cl = cl.g;;
let droite cl = cl.d;;
(* Printing, parsing, etc ... *)
let to_string  = function 
    { g=g; d=d} ->
      (Outils.list_to_string "," atf_to_string g)
      ^" => "^
      (Outils.list_to_string "," atf_to_string d);;
let print cl = print_string(to_string cl);;
let parse =
  parser
      [<g = Lexparse.parse_list "," atf_parse;
	'Genlex.Kwd "=>"; 
	d=Lexparse.parse_list "," atf_parse >] -> faire g d;;
let of_string str = parse(Lexparse.lexer str);;

(* Thories *)
let theorie_faire liste = liste;;
(* Printing, parsing, etc ... *)
let theorie_to_string theorie =  
  "{\n"^
      (Outils.list_to_string ";\n" to_string theorie)
  ^"\n}";;
let theorie_print theorie = print_string(theorie_to_string theorie);;
let theorie_parse =
  parser
      [<'Genlex.Kwd "{";
	clauses = Lexparse.parse_list ";" parse; 'Genlex.Kwd "}">] ->
	clauses ;;
let theorie_of_string str = theorie_parse(Lexparse.lexer str);;

(* Autres fonctions *) 

let length cl = (List.length cl.g,List.length cl.d);;
let apply subst = function
    Formula(p,terms) -> 
      Formula(p, List.map (Terme.apply subst) terms);; 
let apply_to_clause subst = function
    { g=g; d=d } ->
      { g=List.map (apply subst) g; d=List.map (apply subst) d };;

let unify = function
     (Formula(p1,terms1), Formula(p2,terms2)) ->
       if 
	 ( p1 = p2 && (List.length terms1) = (List.length terms2) )
       then
	 Terme.unify (List.combine terms1 terms2 )
       else
	 failwith "unify";;

let  remove n l = 
  let rec remove m acc = function
      [] -> failwith "remove"
    | t::q -> if m = 0 then (List.rev acc) @ q
      else remove (m -1) (t::acc) q
  in
    remove n [] l
;;

(* Predicats pour la simplification *)

(* Est une clause une tautologie ? *)
let taut {g=g;d=d} = 
  let 
    predicat af_d = List.exists (fun af_g -> af_g = af_d) g
  in
    List.exists predicat d;;

let filtrer = function
    (Formula(p1,terms1), Formula(p2,terms2)) ->
      if 
	(p1 = p2 && (List.length terms1) = (List.length terms2) )
      then
	Terme.filtrer (List.combine terms1 terms2 )
      else
	failwith "filtrer";;

(* Est une clause une sous clause d'une autre ? *)
(* Plus fin de la logique propositionelle, *)
(* Il faut se servir du filtrage. *)

let rec filtrer_l at liste = 
  match liste with
      [] -> []
    | t::q -> 
	let
	  tetel =
	  try
	    [ filtrer (at,t) ]
	  with
	      Failure "filtrer" -> []
	in
	  tetel @ (filtrer_l at q);;

let rec sous_l liste1 liste2 = match liste1 with
    [] -> [Terme.identite]
  | t::q -> 
      let
	candidatq =  
	sous_l q liste2
      and
	candidatt =  filtrer_l t liste2
      in
	Terme.merge_l candidatt candidatq ;;
	
	
let sous {g=g1;d=d1} {g=g2;d=d2} =
  let
    listeg = sous_l g1 g2
  and
    listed = sous_l d1 d2
  in
  let
    candidats = Terme.merge_l listeg listed
  in
    if candidats = [] then false else true;;

