type justification = Ax | Fact of int*int | Res of int*int*int*int;;
type ('a,'b,'c) jclause = 
    { clause : ('a,'b,'c) Clause.clause;
      no : int;
      regle : justification };;

type ('a,'b,'c) preuve = ('a,'b,'c) jclause list;;

let gl_compteur=ref 0;;
let incr compteur = (compteur:= 1 + !compteur);;
 
(* Construction/deconstruction des objets *)
let of_theorie theorie =
  gl_compteur := 0;
  (List.map 
     (fun cl ->
	let 
	  result =
	  { clause=cl;
	    no=(!gl_compteur);
	    regle=Ax }
	in
	  incr(gl_compteur);
	  result
     )
     theorie
  );;

let clause jcl = jcl.clause;;

let jclause_to_string {clause = c; no = n; regle = r} =
  let
    strregle = match r with
	Ax -> "Axiome"
      | Fact(i,j) -> "Fact "^
	(string_of_int i)^" "^
	(string_of_int j)
      | Res(g,i,d,j) -> 
	  "Res "^
	  (string_of_int g)^"."^(string_of_int i)^";"^
	  (string_of_int d)^"."^(string_of_int j)
  in
    
    (string_of_int n)^" :\n"^
    (Clause.to_string c)^"\n"^
    "\t\t\tpar : "^strregle ;;

let to_string preuve =
	  (Outils.list_to_string ";\n" jclause_to_string preuve)
	  

let print preuve = print_string(to_string preuve);;

(** Le calcul R **)

let rec factoriser (jc,i,j) = 
  let
    g = Clause.gauche(jc.clause)
  and
    d = Clause.droite(jc.clause)
  in
  let
    f_at1 = List.nth d i 
  and
    f_at2 = List.nth d j 
  in
    try
      let
	sigma = Clause.unify (f_at1,f_at2)
      in
      let
	cl = Clause.apply_to_clause sigma
		   (Clause.faire g (Clause.remove i d))
      in
      let
	result = {clause=cl;no=(!gl_compteur);regle= Fact(jc.no,i)}
		   
      in
	incr(gl_compteur); result
    with
	Failure "unify" -> failwith "factoriser"
;;

let rec resolution (jcg,i) (jcd,j) =
  let 
    gd = Clause.droite(jcg.clause)
  and
    dg = Clause.gauche(jcd.clause)
  in
  let
    f_at1 = List.nth gd i 
  and
    f_at2 = List.nth dg j 
  in
    try
      let
	sigma = Clause.unify (f_at1,f_at2)
      in
      let
	gg = Clause.gauche(jcg.clause)
      and
  	dd = Clause.droite(jcd.clause)
      in
      let
	cl = Clause.apply_to_clause sigma 
	       (Clause.faire 
		  (gg@(Clause.remove j dg) ) 
		  ((Clause.remove i gd)@dd ))
      in
      let
	result =
	{
	  clause=cl;no=(!gl_compteur);
	  regle=Res(jcg.no,i,jcd.no,j)
	}
      in
	incr(gl_compteur); result		 
    with
	Failure "unify" -> failwith "resolution"
;;

(* Autres operations *)

(* Il y a t'il une contradiction ? *)
let rec contradiction = function
    [] -> false
  | ({clause=cl;no=_;regle=_}::q) -> 
      if cl = Clause.vide then true 
      else
	(contradiction q);;

let sous jc1 jc2 = Clause.sous (clause jc1) (clause jc2);;
