(** Termes **)
type ('a,'b) t = Var of 'b
		 | Operation of 'a * ('a,'b) t list ;;

(* Construction des termes *)
let faire_op a fils = Operation(a,fils);;
let faire_var b  = Var(b);;

(* Printing, parsing, etc. *) 
let rec to_string = function 
    Var(x) -> x
  | Operation(f,fils) ->
      f^"("^(Outils.list_to_string "," to_string fils)^")";;
let print t = print_string("\n"^to_string t^"\n");; 
let parse = 
  let rec 
    parse_t = 
    parser [<'Genlex.Ident x ; fils = parse_f>] ->
      (match fils with
	   None -> faire_var x
	 | Some(f) -> faire_op x f)
  and
    parse_f stream = 
    if (Stream.peek stream) = Some(Genlex.Kwd "(") then
      match stream with
	  parser [<'Genlex.Kwd "(" ; e = Lexparse.parse_list "," parse_t; 'Genlex.Kwd ")">] 
	      -> Some(e)
    else
      match stream with
	  parser [<>] -> None
  in
    parse_t;;
let of_string str = parse(Lexparse.lexer str);;

(* Autres fonctions *)
let rec vars = function 
    Var x -> [x]
  | Operation(f,fils) -> List.concat (List.map vars fils);;


(** Substitutions **) 
type ('a,'b) subst = ('b * ('a,'b) t) list;;

let identite = [];; 
let faire_subst x  t  = [(x,t)] ;;

(*  Printing, parsing, etc. *) 
let subst_to_string subst = 
  let 
    l = List.map (fun (x,t) -> " "^x^" -> "^(to_string t)^" ") subst
  in
    "{\n"^(Outils.list_to_string "\n" (fun x ->x) l)^"\n}\n";;
let print_subst  subst = print_string (subst_to_string subst);;

(* Autres fonctions *)    
let rec apply subst = function 
    Var x -> 
      (
	try 
	  List.assoc x subst
	with
	    Not_found -> Var x
      )
  | Operation(f,fils) -> 
      Operation(f,List.map (apply subst) fils);; 

(* Composition type ; - et non o ; d'abord sub1 et puis s2 *)
let compose sub1 sub2 = 
  let
    predicat = fun (x,t) -> not (List.mem x (List.map fst sub1) )
  in
    (List.map (fun (x,t) -> (x, apply sub2 t)) sub1)
    @ List.filter predicat sub2;;



(** Unification **)

let rec unify = function 
    [] -> identite 
  | tete::queue -> 
      let 
	phi = 
	match tete  with
            (Var x, t ) | (t, Var x ) -> 
	      if List.mem x (vars t) && not (t = Var(x))
	      then failwith "unify"
	      else faire_subst x t
	  | (Operation(f,ff),Operation(g,fg)) ->
	      if( f = g && (List.length ff) = (List.length fg))
              then unify (List.combine ff fg)
              else failwith "unify"
      in
      let 
	psi = 
	unify 
	  (List.map 
	     (fun (t1,t2) -> (apply phi t1, apply phi t2)) queue )
      in
	(* Attention ici  l'ordre de la composition *)
	compose phi psi 
;;

(** Filtrage **)
let add couple subst = 
  match couple with
      (x,t) -> 
	try 
	  if (t = (List.assoc x subst) ) then subst else failwith "add"
	with
	    Not_found -> couple::subst
	  | Failure "add" -> failwith "add";;

let rec merge (sub1: ('a,'b) subst) sub2 = 
  try
    match sub1 with
	[] -> sub2
      | t::q -> merge q (add t sub2)
  with
      Failure "add" -> failwith "merge";;

let merge_el sub liste =
  Outils.mapfilter (fun t -> merge sub t ) liste;;

let merge_l liste1 liste2 =
  List.concat (List.map (fun sub -> merge_el sub liste2) liste1);;

(* Voici le filtrage *)
let rec filtrer = function
  [] -> identite 
  | tete::queue -> 
      let 
	phi = match tete  with
            (Var x, t )  -> 
	      if List.mem x (vars t) then failwith "filtrer"
	      else faire_subst x t
	  | (Operation(f,ff),Operation(g,fg)) ->
	      if( f = g && (List.length ff) = (List.length fg))
              then filtrer (List.combine ff fg)
              else failwith "filtrer"
	  | (_,Var x)  -> failwith "filtrer"
      and
	psi = filtrer queue
      in
	try 
	  merge psi phi
	with
	    Failure "merge" -> failwith "filtrer"
;;
      
